@@ -20,6 +20,7 @@ from morango.constants.capabilities import ASYNC_OPERATIONS
 from morango.constants import transfer_stages
 from morango.constants import transfer_statuses
 from morango.errors import MorangoResumeSyncError
+from morango.errors import MorangoLimitExceeded
 from morango.models.certificates import Filter
 from morango.models.core import Buffer
 from morango.models.core import DatabaseMaxCounter
@@ -55,6 +56,8 @@ if "postgresql" in transaction.get_connection(USING_DB).vendor:
             which copies all the configuration settings of the `default` db connection",
     )
 
+SQL_UNION_MAX = 500
+
 
 class OperationLogger(object):
     def __init__(self, start_msg, end_msg):
@@ -413,7 +416,6 @@ def _queue_into_buffer(transfersession):
     We use raw sql queries to place data in the buffer and the record max counter buffer, which matches the conditions of the FSIC,
     as well as the partition for the data we are syncing.
     """
-    last_saved_by_conditions = []
     filter_prefixes = Filter(transfersession.filter)
     server_fsic = json.loads(transfersession.server_fsic)
     client_fsic = json.loads(transfersession.client_fsic)
@@ -427,18 +429,7 @@ def _queue_into_buffer(transfersession):
     if not fsics:
         return
 
-    # create condition for all push FSICs where instance_ids are equal, but internal counters are higher than FSICs counters
-    for instance, counter in six.iteritems(fsics):
-        last_saved_by_conditions += [
-            "(last_saved_instance = '{0}' AND last_saved_counter > {1})".format(
-                instance, counter
-            )
-        ]
-    if fsics:
-        last_saved_by_conditions = [
-            _join_with_logical_operator(last_saved_by_conditions, "OR")
-        ]
-
+    profile_condition = ["profile = '{}'".format(transfersession.sync_session.profile)]
     partition_conditions = []
     # create condition for filtering by partitions
     for prefix in filter_prefixes:
@@ -446,47 +437,98 @@ def _queue_into_buffer(transfersession):
     if filter_prefixes:
         partition_conditions = [_join_with_logical_operator(partition_conditions, "OR")]
 
-    # combine conditions
-    fsic_and_partition_conditions = _join_with_logical_operator(
-        last_saved_by_conditions + partition_conditions, "AND"
-    )
+    chunk_size = 200
+    fsics = list(fsics.items())
+    fsics_len = len(fsics)
+    fsics_limit = chunk_size * SQL_UNION_MAX
 
-    # filter by profile
-    where_condition = _join_with_logical_operator(
-        [
-            fsic_and_partition_conditions,
-            "profile = '{}'".format(transfersession.sync_session.profile),
-        ],
-        "AND",
-    )
+    if fsics_len >= fsics_limit:
+        raise MorangoLimitExceeded(
+            "Limit of {limit} instance counters exceeded with {actual}".format(
+                limit=fsics_limit, actual=fsics_len
+            )
+        )
 
-    # execute raw sql to take all records that match condition, to be put into buffer for transfer
-    with connection.cursor() as cursor:
-        queue_buffer = """INSERT INTO {outgoing_buffer}
-                        (model_uuid, serialized, deleted, last_saved_instance, last_saved_counter, hard_deleted,
-                         model_name, profile, partition, source_id, conflicting_serialized_data, transfer_session_id, _self_ref_fk)
-                        SELECT id, serialized, deleted, last_saved_instance, last_saved_counter, hard_deleted, model_name, profile, partition, source_id, conflicting_serialized_data, '{transfer_session_id}', _self_ref_fk
-                        FROM {store} WHERE {condition}""".format(
-            outgoing_buffer=Buffer._meta.db_table,
-            transfer_session_id=transfersession.id,
-            condition=where_condition,
-            store=Store._meta.db_table,
+    # chunk fsics creating multiple SQL selects which will be unioned before insert
+    i = 0
+    chunk = fsics[:chunk_size]
+    select_buffers = []
+    select_rmc_buffers = []
+
+    while chunk:
+        # create condition for all push FSICs where instance_ids are equal, but internal counters are higher than
+        # FSICs counters
+        last_saved_by_conditions = [
+            "(last_saved_instance = '{0}' AND last_saved_counter > {1})".format(
+                instance, counter
+            )
+            for instance, counter in chunk
+        ]
+        if last_saved_by_conditions:
+            last_saved_by_conditions = [
+                _join_with_logical_operator(last_saved_by_conditions, "OR")
+            ]
+
+        # combine conditions and filter by profile
+        where_condition = _join_with_logical_operator(
+            profile_condition + last_saved_by_conditions + partition_conditions, "AND"
+        )
+
+        # execute raw sql to take all records that match condition, to be put into buffer for transfer
+        select_buffers.append(
+            """SELECT
+                   id, serialized, deleted, last_saved_instance, last_saved_counter, hard_deleted, model_name, profile,
+                   partition, source_id, conflicting_serialized_data,
+                   CAST ('{transfer_session_id}' AS {transfer_session_id_type}), _self_ref_fk
+               FROM {store} WHERE {condition}
+            """.format(
+                transfer_session_id=transfersession.id,
+                transfer_session_id_type=TransferSession._meta.pk.rel_db_type(
+                    connection
+                ),
+                condition=where_condition,
+                store=Store._meta.db_table,
+            )
         )
-        cursor.execute(queue_buffer)
         # take all record max counters that are foreign keyed onto store models, which were queued into the buffer
-        queue_rmc_buffer = """INSERT INTO {outgoing_rmcb}
-                            (instance_id, counter, transfer_session_id, model_uuid)
-                            SELECT instance_id, counter, '{transfer_session_id}', store_model_id
-                            FROM {record_max_counter} AS rmc
-                            INNER JOIN {outgoing_buffer} AS buffer ON rmc.store_model_id = buffer.model_uuid
-                            WHERE buffer.transfer_session_id = '{transfer_session_id}'
-                            """.format(
-            outgoing_rmcb=RecordMaxCounterBuffer._meta.db_table,
-            transfer_session_id=transfersession.id,
-            record_max_counter=RecordMaxCounter._meta.db_table,
-            outgoing_buffer=Buffer._meta.db_table,
+        select_rmc_buffers.append(
+            """SELECT instance_id, counter, CAST ('{transfer_session_id}' AS {transfer_session_id_type}), store_model_id
+               FROM {record_max_counter} AS rmc
+               INNER JOIN {outgoing_buffer} AS buffer ON rmc.store_model_id = buffer.model_uuid
+               WHERE buffer.transfer_session_id = '{transfer_session_id}'
+            """.format(
+                transfer_session_id=transfersession.id,
+                transfer_session_id_type=TransferSession._meta.pk.rel_db_type(
+                    connection
+                ),
+                record_max_counter=RecordMaxCounter._meta.db_table,
+                outgoing_buffer=Buffer._meta.db_table,
+            )
+        )
+        i += chunk_size
+        chunk = fsics[i : i + chunk_size]
+
+    with connection.cursor() as cursor:
+        cursor.execute(
+            """INSERT INTO {outgoing_buffer}
+               (model_uuid, serialized, deleted, last_saved_instance, last_saved_counter,
+               hard_deleted, model_name, profile, partition, source_id, conflicting_serialized_data,
+               transfer_session_id, _self_ref_fk)
+               {select}
+            """.format(
+                outgoing_buffer=Buffer._meta.db_table,
+                select=" UNION ".join(select_buffers),
+            )
+        )
+        cursor.execute(
+            """INSERT INTO {outgoing_rmcb}
+               (instance_id, counter, transfer_session_id, model_uuid)
+               {select}
+            """.format(
+                outgoing_rmcb=RecordMaxCounterBuffer._meta.db_table,
+                select=" UNION ".join(select_rmc_buffers),
+            )
         )
-        cursor.execute(queue_rmc_buffer)
 
 
 @transaction.atomic(using=USING_DB)
@@ -591,9 +633,7 @@ class InitializeOperation(LocalOperation):
         # attributes that we'll use to identify existing sessions. we really only want there to
         # be one of these at a time
         data = dict(
-            push=context.is_push,
-            sync_session_id=context.sync_session.id,
-            active=True,
+            push=context.is_push, sync_session_id=context.sync_session.id, active=True
         )
 
         # get the most recent transfer session
@@ -650,13 +690,12 @@ class SerializeOperation(LocalOperation):
         self._assert(context.filter is not None)
 
         if context.is_producer and SETTINGS.MORANGO_SERIALIZE_BEFORE_QUEUING:
-            _serialize_into_store(
-                context.sync_session.profile,
-                filter=context.filter,
-            )
+            _serialize_into_store(context.sync_session.profile, filter=context.filter)
 
         fsic = json.dumps(
-            DatabaseMaxCounter.calculate_filter_max_counters(context.filter)
+            DatabaseMaxCounter.calculate_filter_specific_instance_counters(
+                context.filter, is_producer=context.is_producer
+            )
         )
         if context.is_server:
             context.transfer_session.server_fsic = fsic
@@ -824,14 +863,8 @@ class ReceiverDeserializeOperation(LocalOperation):
         records_transferred = context.transfer_session.records_transferred or 0
         if SETTINGS.MORANGO_DESERIALIZE_AFTER_DEQUEUING and records_transferred > 0:
             # we first serialize to avoid deserialization merge conflicts
-            _serialize_into_store(
-                context.sync_session.profile,
-                filter=context.filter,
-            )
-            _deserialize_from_store(
-                context.sync_session.profile,
-                filter=context.filter,
-            )
+            _serialize_into_store(context.sync_session.profile, filter=context.filter)
+            _deserialize_from_store(context.sync_session.profile, filter=context.filter)
 
         # update database max counters but use latest fsics from client/server
         if context.is_receiver:
@@ -1013,6 +1046,10 @@ class LegacyNetworkInitializeOperation(NetworkOperation):
         data = self.create_transfer_session(context)
         context.transfer_session.server_fsic = data.get("server_fsic") or "{}"
 
+        # A legacy instance performs queuing during the creation of the transfer session, so since we use a new
+        # workflow we need to update the network server when pushing to say how many records we've queued. For pull,
+        # we handle that here in the initialization/creation of the transfer session,
+        # since that's when it's first available.
         if context.transfer_session.pull:
             context.transfer_session.records_total = data.get("records_total", 0)
 
@@ -1094,7 +1131,19 @@ class LegacyNetworkQueueOperation(NetworkLegacyNoOpMixin, NetworkOperation):
     Without ASYNC_OPERATIONS capability, the server will perform queuing during initialization
     """
 
-    pass
+    def handle(self, context):
+        """
+        :type context: NetworkSessionContext
+        """
+        self._assert(ASYNC_OPERATIONS not in context.capabilities)
+
+        # A legacy instance performs queuing during the creation of the transfer session, so since we use a new
+        # workflow we need to update the network server when pushing to say how many records we've queued. For pull,
+        # we handle that in the initialization/creation of the transfer session, since that's when it's first available.
+        if context.is_push:
+            self.update_transfer_session(context, records_total=context.transfer_session.records_total)
+
+        return transfer_statuses.COMPLETED
 
 
 class NetworkQueueOperation(NetworkOperation):
@@ -1204,7 +1253,7 @@ class NetworkPullTransferOperation(NetworkOperation):
         return op_status
 
 
-class LegacyDequeueOperation(NetworkLegacyNoOpMixin, NetworkOperation):
+class LegacyNetworkDequeueOperation(NetworkLegacyNoOpMixin, NetworkOperation):
     """
     Without ASYNC_OPERATIONS capability, the server will perform dequeuing during cleanup
     """
